==> dclib-0.3.14a.tar.bz2: dclib-0.3.14a/dclib/cfilehasher.cpp <==
/***************************************************************************
        cfilehasher.h  -  Calculate the TTH root and leaves for a file
                             -------------------
    begin                : Fri May 16 2008
    copyright            : (C) 2008 by Edward Sheldrake
    email                : ejs1920@yahoo.co.uk
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cfilehasher.h"
 
#include "core/cbase32.h"
#include "core/cdir.h"

#include "hash/compat.h"
#include "hash/MerkleTree.h"

#include <stdio.h>
#include <string.h> //for memcmp()

/** */
CString CTreeVerificationReport::ToString() const
{
	CString s;
	
	s += "CTreeVerificationReport\n";
	s += "FileName: " + filename + "\n";
	s += "FileSize: " + CString().setNum(filesize) + "\n";
	s += "Actual TTH root: " + tthRoot + "\n";
	s += "All hashes OK: ";
	if ( allgood )
	{
		s += "YES";
	}
	else
	{
		s += "NO";
	}
	s += "\n";
	s += "Number of segments: " + CString().setNum(segments->Count()) + "\n";
	s += "\n";
	//    18446744073709551615 18446744073709551615 ORQ6K7F5U2QGJOUKEKPEDJ2XXM3FTVYQ2BLWQOI ORQ6K7F5U2QGJOUKEKPEDJ2XXM3FTVYQ2BLWQOI
	s += "Start                Size                 Expected TTH of block                   Actual TTH of block\n";
	
	CHashedSegment * segment = 0;
	
	while ( (segment = segments->Next(segment)) != 0 )
	{
		s += CString().setNum(segment->start).RightJustify(20);
		s += " ";
		s += CString().setNum(segment->size).RightJustify(20);
		s += " ";
		s += segment->expected;
		s += " ";
		s += segment->actual;
		
		if ( segment->expected != segment->actual )
		{
			s += " *** MISMATCH ***";
		}
		
		s += "\n";
	}
	
	return s;
}

/** */
CFileHasher::CFileHasher( const CString filename, CByteArray * workmem )
{
	status      = efhsNotStarted;
	m_bStop     = FALSE;
	m_pRootData = 0;
	m_pLeafData = 0;
	filesize    = 0;
	m_sProgress = "Initialising...";
	
	if ( file.Open( filename, IO_RAW | IO_READONLY ) == TRUE )
	{
		m_pWorkMem = workmem;
		usingOwnMem = FALSE;
		filesize = CDir().getFileSize( filename, FALSE );
		
		if ( m_pWorkMem == 0 )
		{
			m_pWorkMem = new CByteArray( 1024*1024 );
			usingOwnMem = TRUE;
		}
		
		if ( m_pWorkMem == 0 )
		{
			printf( "CFileHasher memory allocation failure\n" );
			status = efhsError;
		}
		else
		{
			m_sProgress = "Ready.";
			status = efhsReady;
		}
	}
	else
	{
		m_pWorkMem = 0;
		printf( "CFileHasher cannot open '%s'\n", filename.Data() );
		status = efhsError;
	}
}

/** delete stuff, close file */
CFileHasher::~CFileHasher()
{
	if ( m_pRootData != 0 )
	{
		delete m_pRootData;
		m_pRootData = 0;
	}
	
	if ( m_pLeafData != 0 )
	{
		delete m_pLeafData;
		m_pLeafData = 0;
	}
	
	if ( (usingOwnMem == TRUE) && (m_pWorkMem != 0) )
	{
		delete m_pWorkMem;
		m_pWorkMem = 0;
	}
	
	if ( file.IsOpen() )
	{
		file.Close();
	}
}

/** */
CString CFileHasher::GetHashRoot()
{
	CString s;
	
	if ( GetStatus() == efhsFinished )
	{
		if ( m_pRootData != 0 )
		{
			if ( m_pRootData->Size() == TigerTree::HASH_SIZE )
			{
				CBase32 base32;
				CByteArray result;
				base32.Encode( &result, m_pRootData );
				if ( result.Size() > 0 )
				{
					s.Set( (const char*)result.Data(), result.Size() );
				}
				else
				{
					printf("CFileHasher::GetHashRoot base32 encode error\n");
				}
			}
			else
			{
				printf("CFileHasher::GetHashRoot wrong size %lu\n", m_pRootData->Size());
			}
		}
		else
		{
			printf("CFileHasher::GetHashRoot m_pRootData==0\n");
		}
	}
	else
	{
		printf("CFileHasher::GetHashRoot not finished\n");
	}
	
	return s;
}

/** */
CByteArray * CFileHasher::GetHashRootRaw()
{
	CByteArray * result = 0;
	
	if ( GetStatus() == efhsFinished )
	{
		if ( m_pRootData != 0 )
		{
			if ( m_pRootData->Size() == TigerHash::HASH_SIZE )
			{
				result = new CByteArray();
				result->Append( m_pRootData->Data(), m_pRootData->Size() );
			}
			else
			{
				printf("CFileHasher::GetHashRootRaw wrong size %lu\n", m_pRootData->Size());
			}
		}
		else
		{
			printf("CFileHasher::GetHashRootRaw m_pRootData==0\n");
		}
	}
	else
	{
		printf("CFileHasher::GetHashRootRaw not finished\n");
	}
	
	return result;
}

/** */
CByteArray * CFileHasher::GetLeafData()
{
	CByteArray * result = 0;
	
	if ( GetStatus() == efhsFinished )
	{
		if ( m_pLeafData != 0 )
		{
			result = new CByteArray();
			result->Append( m_pLeafData->Data(), m_pLeafData->Size() );
		}
		else
		{
			printf("CFileHasher::GetLeafData m_pLeafData==0\n");
		}
	}
	else
	{
		printf("CFileHasher::GetLeafData not finished\n");
	}
	
	return result;
}

/** */
void CFileHasher::StopHashing()
{
	m_LockThread.Lock();
	
	m_bStop = TRUE;
	
	m_LockThread.UnLock();
}

/** */
void CFileHasher::SetStatus( const eFileHasherStatus newstatus )
{
	m_LockThread.Lock();
	
	status = newstatus;
	
	m_LockThread.UnLock();
}

/** */
void CFileHasher::SetProgress( const CString s )
{
	m_LockThread.Lock();
	
	m_sProgress = s;
	
	m_LockThread.UnLock();
}

/** */
void CFileHasher::ComputeHash( const eFileHasherStatus endStatus, const CString progressPrefix )
{
	if ( GetStatus() == efhsReady )
	{
		long filepos = 0;
		
		SetStatus( efhsWorking );
		
		TigerTree hasher( max(TigerTree::calcBlockSize(filesize,10), 64*1024 ) );
		
		long len;
		
		while ( ((len = file.Read( (char*)m_pWorkMem->Data(), m_pWorkMem->Size() )) > 0) && (IsStop() == FALSE) )
		{
			hasher.update(m_pWorkMem->Data(),len);

			filepos += len;
			SetProgress( progressPrefix + CString().setNum(filepos) + "/" + CString().setNum(filesize) );
		}
		
		file.Close();
		
		if ( IsStop() == TRUE )
		{
			SetStatus( efhsError );
		}
		else
		{
			hasher.finalize();
			
			m_pRootData = new CByteArray();
			m_pRootData->Append( hasher.getRoot().data, TigerHash::HASH_SIZE );
			
			vector<uint8_t> leafdata = hasher.getLeafData();
			m_pLeafData = new CByteArray();
			m_pLeafData->Append( &leafdata[0], leafdata.size() );
		
			SetStatus( endStatus );
		}
		
		if ( usingOwnMem == TRUE )
		{
			delete m_pWorkMem;
			m_pWorkMem = 0;
		}
	}
	else
	{
		printf("CFileHasher::ComputeHash not ready\n");
	}
}

/** from DownloadManager::endData in the DC++ source */
int64_t CFileHasher::GetBlockSize( const unsigned long leavesSize, const int64_t filesize )
{
	int64_t blocksize = 1024;
	while ( blocksize * leavesSize/sizeof(TigerTree::MerkleValue) < filesize )
	{
		blocksize = blocksize * 2;
	}
	return blocksize;
}

/** */
bool CFileHasher::ValidateHashLeaves( CString tth, const CByteArray * leaves, const ulonglong filesize )
{
	CBase32 base32;
	CByteArray dst;
	
	int size = base32.Decode( &dst, &tth );
	
	if ( size != TigerTree::HASH_SIZE )
	{
		printf("CFileHasher::ValidateHashLeaves: base32 decode return wrong size %d\n",size);
		return FALSE;
	}
	else
	{
		return ValidateHashLeaves( &dst, leaves, filesize );
	}
}

/** */
bool CFileHasher::ValidateHashLeaves( const CByteArray * root, const CByteArray * leaves, const ulonglong filesize )
{
	if ( (root == 0) || (leaves == 0) )
	{
		printf("CFileHasher::ValidateHashLeaves null pointer\n");
		return FALSE;
	}
	
	CByteArray rootcopy;
	rootcopy.Append(((CByteArray*)root)->Data(),root->Size());
	
	CByteArray leavescopy;
	leavescopy.Append(((CByteArray*)leaves)->Data(),leaves->Size());
	
	if (rootcopy.Size() != TigerTree::HASH_SIZE )
	{
		printf("CFileHasher::ValidateHashLeaves root wrong size\n");
		return FALSE;
	}
	else if (leavescopy.Size() < TigerTree::HASH_SIZE )
	{
		printf("CFileHasher::ValidateHashLeaves leaves too small\n");
		return FALSE;
	}
	else if ( leavescopy.Size() == TigerTree::HASH_SIZE ) // optimisation
	{
		if ( memcmp(rootcopy.Data(),leavescopy.Data(),TigerTree::HASH_SIZE) == 0 )
		{
			return TRUE;
		}
	}
	
	int64_t size = (int64_t) filesize;
	
	if ( size < 0 )
	{
		printf("CFileHasher::ValidateHashLeaves filesize too big\n");
		return FALSE;
	}
	
	int64_t blocksize = GetBlockSize( leavescopy.Size(), size );
	
	TigerTree hasher(size,blocksize,leavescopy.Data());
	
	if ( memcmp(hasher.getRoot().data,rootcopy.Data(),TigerTree::HASH_SIZE) == 0 )
	{
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

/** */
CByteArray * CFileHasher::HashByteArray( CByteArray * data, unsigned long length )
{
	CByteArray * result = new CByteArray();
	
	TigerTree tth;
	tth.update( data->Data(), length );
	tth.finalize();
	result->Append( tth.getRoot().data, TigerTree::HASH_SIZE );
	
	return result;
}

/** */
void CFileHasherThread::Thread( CObject * )
{
	Lock();
	
	ComputeHash();
	
	UnLock();
	
	Stop();
}

/** */
CFileTreeVerifier::CFileTreeVerifier( const CString filename, CByteArray * leaves, CByteArray * workmem ) : CFileHasher ( filename, workmem )
{
	m_pLeaves = leaves;
	m_pReport = new CTreeVerificationReport();
	m_pReport->filename = filename;
}

/** */
CFileTreeVerifier::~CFileTreeVerifier()
{
	if ( m_pReport != 0 )
	{
		delete m_pReport;
		m_pReport = 0;
	}
}

/** */
CTreeVerificationReport * CFileTreeVerifier::GetReport()
{
	CTreeVerificationReport * pointer = 0;
	
	if ( GetStatus() == efhsFinished )
	{
		pointer = m_pReport;
		m_pReport = 0;
	}
	
	return pointer;
}

/** */
void CFileTreeVerifier::Thread( CObject * )
{
	Lock();
	
	SetProgress( "Initialising..." );
	ComputeHash( efhsWorking, "PASS1 " );
	
	if ( GetStatus() == efhsWorking )
	{
		CBase32 base32;
		bool ok = TRUE;
		
		/* first get segment size, we need it for either method */
		int leafcount = m_pLeaves->Size() / TigerTree::HASH_SIZE;
		unsigned long blocksize = 1024;
		while ( blocksize * leafcount < filesize )
		{
			blocksize = blocksize * 2;
		}
			
		/* Can we do this? Or could two lists of tth leaves the same length be from differently structured trees? */
		/* This should not take to long, no need to check m_bStop */
		/* if ( m_pLeaves->Size() == HashLeavesDirect()->Size() )
		{
			printf("CFileTreeVerifier: leaf data are same length, comparing\n");
			CByteArray expected_in;
			CByteArray actual_in;
			CByteArray expected_out;
			CByteArray actual_out;
			
			for ( unsigned long i = 0; i < m_pLeaves->Size(); i = i + TigerTree::HASH_SIZE )
			{
				expected_in.SetSize(0);
				actual_in.SetSize(0);
				expected_out.SetSize(0);
				actual_out.SetSize(0);
				
				expected_in.Append( m_pLeaves->Data()+i, TigerTree::HASH_SIZE );
				actual_in.Append( HashLeavesDirect()->Data()+i, TigerTree::HASH_SIZE );
				
				base32.Encode( &expected_out, &expected_in );
				base32.Encode( &actual_out, &actual_in );
				
				if ( memcmp(expected_in->Data(),expected_out->Data(),TigerTree::HASH_SIZE) != 0 )
				{
					ok = FALSE;
				}
				
				CHashedSegment * segment = new CHashedSegment();
				segment->expected.Set( (const char*) expected_out.Data(), expected_out.Size() );
				segment->actual.Set( (const char*) actual_out.Data(), actual_out.Size() );
				
				segment->start = i * blocksize;
				segment->size = blocksize;
				if ( segment->start + segment->size > filesize )
				{
					segment->size = filesize - segment->start;
				}
				
				m_pReport->segments->Add(segment);
			}
			
			m_pReport->tthRoot  = GetHashRoot();
			m_pReport->filesize = filesize;
			m_pReport->allgood  = ok;
			SetStatus( efhsFinished );
		}
		else
		{
			printf("CFileTreeVerifier: need to rehash each segment of file\n"); */
			/* Otherwise, we need to hash each segement of the file */
			if ( file.Open( m_pReport->filename, IO_RAW | IO_READONLY ) == TRUE )
			{
				CByteArray expected_in;
				CByteArray expected_out;
				CByteArray actual_out;
				ulonglong filepos = 0;
				
				for ( int i = 0; i < leafcount; i++ )
				{
					expected_in.SetSize(0);
					expected_in.Append( m_pLeaves->Data()+(i*TigerTree::HASH_SIZE), TigerTree::HASH_SIZE );
					
					CByteArray * buffer = new CByteArray( blocksize );
					
					long read = file.Read( (char*) buffer->Data(), blocksize );
					if ( read > 0 )
					{
						CByteArray * actual_in = HashByteArray( buffer, (unsigned long)read );
						
						expected_out.SetSize(0);
						actual_out.SetSize(0);
						
						base32.Encode( &expected_out, &expected_in );
						base32.Encode( &actual_out, actual_in );
						
						delete actual_in;
						
						CHashedSegment * segment = new CHashedSegment();
						segment->expected.Set( (const char*)expected_out.Data(), expected_out.Size() );
						segment->actual.Set( (const char*)actual_out.Data(), actual_out.Size() );
						
						if ( segment->expected != segment->actual )
						{
							ok = FALSE;
						}
						
						segment->start = filepos;
						segment->size = read;
						
						m_pReport->segments->Add(segment);
						
						filepos += read;
						SetProgress( "PASS2 " + CString().setNum(filepos) + "/" + CString().setNum(filesize) );
					}
					else
					{
						SetStatus( efhsError );
					}
	
					delete buffer;

					if ( ( IsStop() ) || ( GetStatus() == efhsError ) )
					{
						SetStatus( efhsError );
						break;
					}
					
				}
				
				file.Close();
				if ( GetStatus() == efhsWorking )
				{
					SetStatus ( efhsFinished );
					m_pReport->tthRoot  = GetHashRoot();
					m_pReport->filesize = filesize;
					m_pReport->allgood  = ok;
				}
			}
			else
			{
				printf("CFileTreeVerifier: error opening file '%s'\n", m_pReport->filename.Data());
				SetStatus( efhsError );
			}
		/*} */
	}
	
	UnLock();
	
	Stop();
}

==> dclib-0.3.17.tar.bz2: dclib-0.3.17/dclib/cfilehasher.cpp <==
/***************************************************************************
        cfilehasher.h  -  Calculate the TTH root and leaves for a file
                             -------------------
    begin                : Fri May 16 2008
    copyright            : (C) 2008 by Edward Sheldrake
    email                : ejs1920@yahoo.co.uk
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cfilehasher.h"
 
#include "core/cbase32.h"
#include "core/cdir.h"

#include "hash/compat.h"
#include "hash/MerkleTree.h"

#include <stdio.h>
#include <string.h> //for memcmp()

/** */
CString CTreeVerificationReport::ToString() const
{
	CString s;
	
	s += "CTreeVerificationReport\nFileName: ";
	s += filename;
	s += "\nFileSize: ";
	s += CString().setNum(filesize);
	s += "\nActual TTH root: ";
	s += tthRoot;
	s += "\nAll hashes OK: ";
	if ( allgood )
	{
		s += "YES";
	}
	else
	{
		s += "NO";
	}
	s += "\nNumber of segments: ";
	s += CString().setNum(segments->Count());
	//        18446744073709551615 18446744073709551615 ORQ6K7F5U2QGJOUKEKPEDJ2XXM3FTVYQ2BLWQOI ORQ6K7F5U2QGJOUKEKPEDJ2XXM3FTVYQ2BLWQOI
	s += "\n\nStart                Size                 Expected TTH of block                   Actual TTH of block\n";
	
	CHashedSegment * segment = 0;
	
	while ( (segment = segments->Next(segment)) != 0 )
	{
		s += CString().setNum(segment->start).RightJustify(20);
		s += ' ';
		s += CString().setNum(segment->size).RightJustify(20);
		s += ' ';
		s += segment->expected;
		s += ' ';
		s += segment->actual;
		
		if ( segment->expected != segment->actual )
		{
			s += " *** MISMATCH ***";
		}
		
		s += "\n";
	}
	
	return s;
}

/** */
CFileHasher::CFileHasher( const CString filename, CByteArray * workmem )
{
	status      = efhsNotStarted;
	m_bStop     = FALSE;
	m_pRootData = 0;
	m_pLeafData = 0;
	filesize    = 0;
	m_sProgress = "Initialising...";
	
	if ( file.Open( filename, IO_RAW | IO_READONLY ) == TRUE )
	{
		m_pWorkMem = workmem;
		usingOwnMem = FALSE;
		filesize = CDir().getFileSize( filename, FALSE );
		
		if ( m_pWorkMem == 0 )
		{
			m_pWorkMem = new CByteArray( 1024*1024 );
			usingOwnMem = TRUE;
		}
		
		if ( m_pWorkMem == 0 )
		{
			printf( "CFileHasher memory allocation failure\n" );
			status = efhsError;
		}
		else
		{
			m_sProgress = "Ready.";
			status = efhsReady;
		}
	}
	else
	{
		m_pWorkMem = 0;
		printf( "CFileHasher cannot open '%s'\n", filename.Data() );
		status = efhsError;
	}
}

/** delete stuff, close file */
CFileHasher::~CFileHasher()
{
	if ( m_pRootData != 0 )
	{
		delete m_pRootData;
		m_pRootData = 0;
	}
	
	if ( m_pLeafData != 0 )
	{
		delete m_pLeafData;
		m_pLeafData = 0;
	}
	
	if ( (usingOwnMem == TRUE) && (m_pWorkMem != 0) )
	{
		delete m_pWorkMem;
		m_pWorkMem = 0;
	}
	
	if ( file.IsOpen() )
	{
		file.Close();
	}
}

/** */
CString CFileHasher::GetHashRoot()
{
	CString s;
	
	if ( GetStatus() == efhsFinished )
	{
		if ( m_pRootData != 0 )
		{
			if ( m_pRootData->Size() == TigerTree::HASH_SIZE )
			{
				CBase32 base32;
				CByteArray result;
				base32.Encode( &result, m_pRootData );
				if ( result.Size() > 0 )
				{
					s.Set( (const char*)result.Data(), result.Size() );
				}
				else
				{
					printf("CFileHasher::GetHashRoot base32 encode error\n");
				}
			}
			else
			{
				printf("CFileHasher::GetHashRoot wrong size %lu\n", m_pRootData->Size());
			}
		}
		else
		{
			printf("CFileHasher::GetHashRoot m_pRootData==0\n");
		}
	}
	else
	{
		printf("CFileHasher::GetHashRoot not finished\n");
	}
	
	return s;
}

/** */
CByteArray * CFileHasher::GetHashRootRaw()
{
	CByteArray * result = 0;
	
	if ( GetStatus() == efhsFinished )
	{
		if ( m_pRootData != 0 )
		{
			if ( m_pRootData->Size() == TigerHash::HASH_SIZE )
			{
				result = new CByteArray();
				result->Append( m_pRootData->Data(), m_pRootData->Size() );
			}
			else
			{
				printf("CFileHasher::GetHashRootRaw wrong size %lu\n", m_pRootData->Size());
			}
		}
		else
		{
			printf("CFileHasher::GetHashRootRaw m_pRootData==0\n");
		}
	}
	else
	{
		printf("CFileHasher::GetHashRootRaw not finished\n");
	}
	
	return result;
}

/** */
CByteArray * CFileHasher::GetLeafData()
{
	CByteArray * result = 0;
	
	if ( GetStatus() == efhsFinished )
	{
		if ( m_pLeafData != 0 )
		{
			result = new CByteArray();
			result->Append( m_pLeafData->Data(), m_pLeafData->Size() );
		}
		else
		{
			printf("CFileHasher::GetLeafData m_pLeafData==0\n");
		}
	}
	else
	{
		printf("CFileHasher::GetLeafData not finished\n");
	}
	
	return result;
}

/** */
void CFileHasher::StopHashing()
{
	m_Mutex.Lock();
	
	m_bStop = TRUE;
	
	m_Mutex.UnLock();
}

/** */
void CFileHasher::SetStatus( const eFileHasherStatus newstatus )
{
	m_Mutex.Lock();
	
	status = newstatus;
	
	m_Mutex.UnLock();
}

/** */
void CFileHasher::SetProgress( const CString s )
{
	m_Mutex.Lock();
	
	m_sProgress = s;
	
	m_Mutex.UnLock();
}

/** */
void CFileHasher::ComputeHash( const eFileHasherStatus endStatus, const CString progressPrefix )
{
	if ( GetStatus() == efhsReady )
	{
		long filepos = 0;
		
		SetStatus( efhsWorking );
		
		TigerTree hasher( max(TigerTree::calcBlockSize(filesize,10), 64*1024 ) );
		
		long len;
		
		while ( ((len = file.Read( (char*)m_pWorkMem->Data(), m_pWorkMem->Size() )) > 0) && (IsStop() == FALSE) )
		{
			hasher.update(m_pWorkMem->Data(),len);

			filepos += len;
			CString p = progressPrefix;
			p += CString().setNum(filepos);
			p += '/';
			p += CString().setNum(filesize);
			SetProgress( p );
		}
		
		file.Close();
		
		if ( IsStop() == TRUE )
		{
			SetStatus( efhsError );
		}
		else
		{
			hasher.finalize();
			
			m_pRootData = new CByteArray();
			m_pRootData->Append( hasher.getRoot().data, TigerHash::HASH_SIZE );
			
			vector<uint8_t> leafdata = hasher.getLeafData();
			m_pLeafData = new CByteArray();
			m_pLeafData->Append( &leafdata[0], leafdata.size() );
		
			SetStatus( endStatus );
		}
		
		if ( usingOwnMem == TRUE )
		{
			delete m_pWorkMem;
			m_pWorkMem = 0;
		}
	}
	else
	{
		printf("CFileHasher::ComputeHash not ready\n");
	}
}

/** from DownloadManager::endData in the DC++ source */
int64_t CFileHasher::GetBlockSize( const unsigned long leavesSize, const int64_t filesize )
{
	int64_t blocksize = 1024;
	while ( blocksize * leavesSize/sizeof(TigerTree::MerkleValue) < filesize )
	{
		blocksize = blocksize * 2;
	}
	return blocksize;
}

/** */
bool CFileHasher::ValidateHashLeaves( CString tth, const CByteArray * leaves, const ulonglong filesize )
{
	CBase32 base32;
	CByteArray dst;
	
	int size = base32.Decode( &dst, &tth );
	
	if ( size != TigerTree::HASH_SIZE )
	{
		printf("CFileHasher::ValidateHashLeaves: base32 decode return wrong size %d\n",size);
		return FALSE;
	}
	else
	{
		return ValidateHashLeaves( &dst, leaves, filesize );
	}
}

/** */
bool CFileHasher::ValidateHashLeaves( const CByteArray * root, const CByteArray * leaves, const ulonglong filesize )
{
	if ( (root == 0) || (leaves == 0) )
	{
		printf("CFileHasher::ValidateHashLeaves null pointer\n");
		return FALSE;
	}
	
	CByteArray rootcopy;
	rootcopy.Append(((CByteArray*)root)->Data(),root->Size());
	
	CByteArray leavescopy;
	leavescopy.Append(((CByteArray*)leaves)->Data(),leaves->Size());
	
	if (rootcopy.Size() != TigerTree::HASH_SIZE )
	{
		printf("CFileHasher::ValidateHashLeaves root wrong size\n");
		return FALSE;
	}
	else if (leavescopy.Size() < TigerTree::HASH_SIZE )
	{
		printf("CFileHasher::ValidateHashLeaves leaves too small\n");
		return FALSE;
	}
	else if ( leavescopy.Size() == TigerTree::HASH_SIZE ) // optimisation
	{
		if ( memcmp(rootcopy.Data(),leavescopy.Data(),TigerTree::HASH_SIZE) == 0 )
		{
			return TRUE;
		}
	}
	
	int64_t size = (int64_t) filesize;
	
	if ( size < 0 )
	{
		printf("CFileHasher::ValidateHashLeaves filesize too big\n");
		return FALSE;
	}
	
	int64_t blocksize = GetBlockSize( leavescopy.Size(), size );
	
	TigerTree hasher(size,blocksize,leavescopy.Data());
	
	if ( memcmp(hasher.getRoot().data,rootcopy.Data(),TigerTree::HASH_SIZE) == 0 )
	{
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

/** */
CByteArray * CFileHasher::HashByteArray( CByteArray * data, unsigned long length )
{
	CByteArray * result = new CByteArray();
	
	TigerTree tth;
	tth.update( data->Data(), length );
	tth.finalize();
	result->Append( tth.getRoot().data, TigerTree::HASH_SIZE );
	
	return result;
}

/** */
void CFileHasherThread::Thread( CObject * )
{
	Lock();
	
	ComputeHash();
	
	UnLock();
	
	Stop();
}

/** */
CFileTreeVerifier::CFileTreeVerifier( const CString filename, CByteArray * leaves, CByteArray * workmem ) : CFileHasher ( filename, workmem )
{
	m_pLeaves = leaves;
	m_pReport = new CTreeVerificationReport();
	m_pReport->filename = filename;
}

/** */
CFileTreeVerifier::~CFileTreeVerifier()
{
	if ( m_pReport != 0 )
	{
		delete m_pReport;
		m_pReport = 0;
	}
}

/** */
CTreeVerificationReport * CFileTreeVerifier::GetReport()
{
	CTreeVerificationReport * pointer = 0;
	
	if ( GetStatus() == efhsFinished )
	{
		pointer = m_pReport;
		m_pReport = 0;
	}
	
	return pointer;
}

/** */
void CFileTreeVerifier::Thread( CObject * )
{
	Lock();
	
	SetProgress( "Initialising..." );
	ComputeHash( efhsWorking, "PASS1 " );
	
	if ( GetStatus() == efhsWorking )
	{
		CBase32 base32;
		bool ok = TRUE;
		
		/* first get segment size, we need it for either method */
		int leafcount = m_pLeaves->Size() / TigerTree::HASH_SIZE;
		unsigned long blocksize = 1024;
		while ( blocksize * leafcount < filesize )
		{
			blocksize = blocksize * 2;
		}
			
		/* Can we do this? Or could two lists of tth leaves the same length be from differently structured trees? */
		/* This should not take to long, no need to check m_bStop */
		/* if ( m_pLeaves->Size() == HashLeavesDirect()->Size() )
		{
			printf("CFileTreeVerifier: leaf data are same length, comparing\n");
			CByteArray expected_in;
			CByteArray actual_in;
			CByteArray expected_out;
			CByteArray actual_out;
			
			for ( unsigned long i = 0; i < m_pLeaves->Size(); i = i + TigerTree::HASH_SIZE )
			{
				expected_in.SetSize(0);
				actual_in.SetSize(0);
				expected_out.SetSize(0);
				actual_out.SetSize(0);
				
				expected_in.Append( m_pLeaves->Data()+i, TigerTree::HASH_SIZE );
				actual_in.Append( HashLeavesDirect()->Data()+i, TigerTree::HASH_SIZE );
				
				base32.Encode( &expected_out, &expected_in );
				base32.Encode( &actual_out, &actual_in );
				
				if ( memcmp(expected_in->Data(),expected_out->Data(),TigerTree::HASH_SIZE) != 0 )
				{
					ok = FALSE;
				}
				
				CHashedSegment * segment = new CHashedSegment();
				segment->expected.Set( (const char*) expected_out.Data(), expected_out.Size() );
				segment->actual.Set( (const char*) actual_out.Data(), actual_out.Size() );
				
				segment->start = i * blocksize;
				segment->size = blocksize;
				if ( segment->start + segment->size > filesize )
				{
					segment->size = filesize - segment->start;
				}
				
				m_pReport->segments->Add(segment);
			}
			
			m_pReport->tthRoot  = GetHashRoot();
			m_pReport->filesize = filesize;
			m_pReport->allgood  = ok;
			SetStatus( efhsFinished );
		}
		else
		{
			printf("CFileTreeVerifier: need to rehash each segment of file\n"); */
			/* Otherwise, we need to hash each segement of the file */
			if ( file.Open( m_pReport->filename, IO_RAW | IO_READONLY ) == TRUE )
			{
				CByteArray expected_in;
				CByteArray expected_out;
				CByteArray actual_out;
				ulonglong filepos = 0;
				
				for ( int i = 0; i < leafcount; i++ )
				{
					expected_in.SetSize(0);
					expected_in.Append( m_pLeaves->Data()+(i*TigerTree::HASH_SIZE), TigerTree::HASH_SIZE );
					
					CByteArray * buffer = new CByteArray( blocksize );
					
					long read = file.Read( (char*) buffer->Data(), blocksize );
					if ( read > 0 )
					{
						CByteArray * actual_in = HashByteArray( buffer, (unsigned long)read );
						
						expected_out.SetSize(0);
						actual_out.SetSize(0);
						
						base32.Encode( &expected_out, &expected_in );
						base32.Encode( &actual_out, actual_in );
						
						delete actual_in;
						
						CHashedSegment * segment = new CHashedSegment();
						segment->expected.Set( (const char*)expected_out.Data(), expected_out.Size() );
						segment->actual.Set( (const char*)actual_out.Data(), actual_out.Size() );
						
						if ( segment->expected != segment->actual )
						{
							ok = FALSE;
						}
						
						segment->start = filepos;
						segment->size = read;
						
						m_pReport->segments->Add(segment);
						
						filepos += read;
						CString p = "PASS2 ";
						p += CString().setNum(filepos);
						p += '/';
						p += CString().setNum(filesize);
						SetProgress( p );
					}
					else
					{
						SetStatus( efhsError );
					}
	
					delete buffer;

					if ( ( IsStop() ) || ( GetStatus() == efhsError ) )
					{
						SetStatus( efhsError );
						break;
					}
					
				}
				
				file.Close();
				if ( GetStatus() == efhsWorking )
				{
					SetStatus ( efhsFinished );
					m_pReport->tthRoot  = GetHashRoot();
					m_pReport->filesize = filesize;
					m_pReport->allgood  = ok;
				}
			}
			else
			{
				printf("CFileTreeVerifier: error opening file '%s'\n", m_pReport->filename.Data());
				SetStatus( efhsError );
			}
		/*} */
	}
	
	UnLock();
	
	Stop();
}
